@@ -37,11 +37,13 @@ from datetime import datetime, timedelta
 
 import adif_io
 
-__VERSION__ = "1.0.2"
+__VERSION__ = "1.1.1"
 
 # merge any calls in the same band and same mode within 115 seconds
 MERGE_WINDOW = 115
 
+# ADIF files are supposed to be strict ascii, but conventionally they are really ISO-8859-1
+ADIF_ENCODING = "latin-1"
 
 # WSJT-X generated fields, minimal output fields as well
 FIELD_ORDER = [
@@ -60,28 +62,116 @@ FIELD_ORDER = [
     'NAME',
 ]
 
-ZONE_FIELDS = ['MY_CQ_ZONE', 'CQZ', 'MY_ITU_ZONE', 'ITUZ']
+# You can't have a QSO without these minimum things (ADIF 3.1.0 spec)
+FIELD_MANDATORY = [
+    'QSO_DATE',
+    'TIME_ON',
+    'CALL',
+    'BAND',
+    'MODE'
+]
+
+# ADIF 3.1.0 specifies field properties
+FIELD_INTEGERS = [
+    'K_INDEX', 'NR_BURSTS', 'NR_PINGS', 'SFI', 'SRX', 'STX']
+FIELD_INTEGERS_POS = [
+    'CQZ', 'FISTS', 'FISTS_CC', 'IOTA_ISLAND_ID', 'ITUZ',
+    'MY_CQ_ZONE', 'MY_FISTS', 'MY_IOTA_ISLAND_ID', 'MY_ITU_ZONE',
+    'TEN_TEN', 'UKMSG']
+FIELD_NUMBERS = [
+    'AGE', 'A_INDEX', 'ANT_AZ', 'ANT_EL', 'DISTANCE', 'FREQ', 'FREQ_RX',
+    'MAX_BURSTS', 'RX_PWR', 'TX_PWR']
+FIELD_ZONES = ['MY_CQ_ZONE', 'CQZ', 'MY_ITU_ZONE', 'ITUZ']
+
+# these are not complete, just the common screwups
+FIELD_MODES = {
+    'DOMINO': ['DOMINOEX', 'DOMINOF'],
+    'JT4': ['JT4A', 'JT4B', 'JT4C', 'JT4D', 'JT4E', 'JT4F', 'JT4G'],
+    'JT65': ['JT65A', 'JT65B', 'JT65B2', 'JT65C', 'JT65C2'],
+    'JT9': ['JT9-1', 'JT9-2', 'JT9-5', 'JT9-10', 'JT9-30',
+            'JT9A', 'JT9B', 'JT9C', 'JT9D', 'JT9E', 'JT9E FAST', 'JT9F', 'JT9F FAST',
+            'JT9G', 'JT9G FAST', 'JT9H', 'JT9H FAST'],
+    'MFSK': ['FSQCALL', 'FT4', 'JS8', 'MFSK4', 'MFSK8', 'MFSK11', 'MFSK16',
+             'MFSK22', 'MFSK31', 'MFSK32', 'MFSK64', 'MFSK128'],
+    'OLIVIA': ['OLIVIA 4/125', 'OLIVIA 4/250', 'OLIVIA 8/250', 'OLIVIA 8/500',
+               'OLIVIA 16/500', 'OLIVIA 16/1000', 'OLIVIA 32/1000'],
+    'PSK': ['FSK31', 'PSK10', 'PSK31', 'PSK63', 'PSK63F', 'PSK125', 'PSK250', 'PSK500',
+            'PSK1000', 'PSKAM10', 'PSKAM31', 'PSKAM50', 'PSKFEC31', 'QPSK31', 'QPSK63',
+            'QPSK125', 'QPSK250', 'QPSK500', 'SIM31'],
+    'QRA64': ['QRA64A, QRA64B, QRA64C, QRA64D, QRA64E'],
+    'RTTY': ['ASCI'],
+    'SSB': ['USB', 'LSB'],
+}
+FIELD_MODES_REVERSE = {
+    submode: mode for mode, submodes in FIELD_MODES.items() for submode in submodes
+}
 
 
-def fixup_qso(qso, path=None):
+class ADIFMalformedQSOError(ValueError):
+    """Malformed ADIF QSO Entry"""
+    def __init__(self, expression, message):
+        self.expression = expression
+        self.message = message
+        super().__init__(message)
+
+
+def fixup_qso_mode(qso):
+    """
+    Some log programs don't follow the ADIF spec on modes and submodes, fix them
+    """
+    mode = qso['MODE']
+    real_mode = FIELD_MODES_REVERSE.get(mode)
+    if real_mode:
+        qso['MODE'] = real_mode
+        qso['SUBMODE'] = mode
+    return qso
+
+
+def fixup_qso(qso, path=""):
     """
     Pre-process an individual QSO record upon load and fix common mistakes.
     """
+    if path:
+        qso['_SOURCE_FILE'] = path
+    # if we're missing a madatory field, mark the QSO and do not process any of it
+    missing_mandatory = {field for field  in FIELD_MANDATORY if field not in qso}
+    if missing_mandatory:
+        qso['_MISSING_FIELDS'] = list(missing_mandatory)
+        raise ADIFMalformedQSOError(qso, "{}: {} missing {}".format(
+            path,
+            "/".join([qso.get(field, field.lower()) for field in FIELD_MANDATORY]),
+            ", ".join(missing_mandatory)))
     for field in qso.keys():
         if isinstance(qso[field], str):
             qso[field] = qso[field].strip()
-    # fixup misreporting of FT4 mode
-    if qso.get('MODE') == "FT4":
-        qso['MODE'] = "MFSK"
-        qso['SUBMODE'] = "FT4"
+    qso = fixup_qso_mode(qso)
     # TX_PWR should only be digits
-    if 'TX_PWR' in qso:
-        if qso['TX_PWR'] == "NaN":
-            del qso['TX_PWR']
-        else:
-            match = re.search(r'(\d+)[Ww]', qso['TX_PWR'])
-            if match:
-                qso['TX_PWR'] = match.group(1)
+    for field in ['TX_PWR', 'RX_PWR']:
+        if field in qso:
+            if qso[field] == "NaN":
+                del qso[field]
+            else:
+                match = re.search(r'([\d\.]+)[Ww]', qso[field])
+                if match:
+                    qso[field] = match.group(1)
+    # if the field is a "PositiveInteger" or "Integer" field, make it an int
+    # some broken logbooks (e.g. HRD) generate Numbers where there should be
+    # Integers--accept them but turn them into ints.
+    for field in FIELD_INTEGERS + FIELD_INTEGERS_POS:
+        if field in qso:
+            qso[field] = int(float(qso[field]))
+    # if the field is a "Number" make it a float, unless it's whole in which case int
+    for field in FIELD_NUMBERS:
+        if field in qso:
+            qso[field] = float(qso[field])
+            if field in ['FREQ', 'FREQ_TX']:
+                # round to 3 digits
+                qso[field] = round(qso[field], 3)
+            else:
+                # leave as an int if possible, otherwise float
+                (part, whole) = math.modf(float(qso[field]))
+                if not part:
+                    qso[field] = int(whole)
     # band should always be uppercase
     for field in ['BAND', 'BAND_RX']:
         if field in qso:
@@ -95,20 +185,10 @@ def fixup_qso(qso, path=None):
         if field in qso:
             qso[field] = "{}{}".format(
                 qso[field][0:4].upper(), qso[field][4:].lower())
-    # the following fields are really integers
-    for field in ['TX_PWR', 'DXCC', 'DISTANCE'] + ZONE_FIELDS:
-        if field in qso:
-            qso[field] = int(qso[field])
-    # the following fields are floats, but should be truncated
-    for field in ['FREQ', 'FREQ_RX']:
-        if field in qso:
-            qso[field] = round(float(qso[field]), 3)
     # remove bad LAT/LON entries
     for field in ['LAT', 'LON']:
         if field in qso and qso[field][1:] == "000 00.000":
             del qso[field]
-    if path:
-        qso['_SOURCE_FILE'] = path
     return qso
 
 
@@ -139,8 +219,8 @@ def merge_dupe_fields(field, first, dupe):
         del dupe[field]
         return
     if field in ['CNTY']:
-        fnslc = first[field].replace(" ", "").lower()
-        dnslc = dupe[field].replace(" ", "").lower()
+        fnslc = first[field].replace(" ", "").casefold()
+        dnslc = dupe[field].replace(" ", "").casefold()
         if fnslc == dnslc:
             # if dupe had spaces, use the one with spaces
             if len(first[field]) < len(dupe[field]):
@@ -153,7 +233,7 @@ def merge_dupe_fields(field, first, dupe):
             del dupe[field]
     if field in ['NAME', 'COMMENT']:
         # prefer mixed case to uppercase only entries
-        if first[field].lower() == dupe[field].lower():
+        if first[field].casefold() == dupe[field].casefold():
             if first[field].isupper():
                 first[field] = dupe[field]
             del dupe[field]
@@ -213,7 +293,7 @@ def merge_qsos(qsos, window):
     """
     buckets = {}
     for qso in sorted(qsos, key=adif_io.time_on):
-        key = "{}_{}_{}".format(qso['CALL'], qso['BAND'], qso['MODE'])
+        key = "{}_{}_{}_{}".format(qso['CALL'], qso['BAND'], qso.get('MODE'), qso.get('SUBMODE'))
         if key not in buckets:
             buckets[key] = []
         buckets[key].append(qso)
@@ -240,7 +320,7 @@ def merge_qsos(qsos, window):
     return sorted(merged_qsos, key=adif_io.time_on)
 
 
-def dump_problems(qsos, path):
+def dump_problems(qsos, malformed, path):
     """
     Report any unmerged fields, break the problem report down both
     by field, and by qso and output the report as a .json file
@@ -263,12 +343,13 @@ def dump_problems(qsos, path):
                         '#SELECTED#': qso[field]
                     }
                 dupe_fields[field]['qsos'][qso_id][source] = dupe[field]
-    if problems:
+    if problems or malformed:
         report = {
             'problems_by_field': dupe_fields,
             'problems_by_qso': problems,
+            'malformed_qsos': malformed,
         }
-        with open(path, "w") as wfd:
+        with open(path, "w", encoding=ADIF_ENCODING) as wfd:
             json.dump(report, wfd, indent=4, sort_keys=True)
 
 
@@ -277,7 +358,7 @@ def adif_write_field(stream, field, entry, comment=""):
     Write a single field out for a QSO in <field:length>[data] format.
     Separate them with spaces.
     """
-    if field in ZONE_FIELDS:
+    if field in FIELD_ZONES:
         entry = "{:02d}".format(int(entry))
     else:
         entry = str(entry)
@@ -365,15 +446,16 @@ def read_adif_file(path) -> list:
     Attempt to read and process an ADIF file and return all of the
     QSO information.
 
-    This is complicated by the fact that LoTW files may not be unicode
-    encoded, and might be ISO-8859-15. First attempt to read the file
-    using unicode, should that fail, revert to character detection.
+    This is complicated by the fact that ADIF files are defined to be
+    ascii only but nobody follows that convention, so they may be latin-1,
+    windows cp1282, or unicode UTF-8 encoded.
     """
     try:
-        with open(path) as adif_file:
+        with open(path, encoding=ADIF_ENCODING) as adif_file:
             adif_string = adif_file.read()
-    except UnicodeDecodeError:
-        with open(path, encoding="latin-1") as adif_file:
+    except ValueError:
+        logging.warning("%s: failed to read using latin-1 encoding, retrying as unicode", path)
+        with open(path, encoding="utf-8") as adif_file:
             adif_string = adif_file.read()
     return adif_io.read_from_string(adif_string)
 
@@ -385,37 +467,51 @@ def main():
     parser = argparse.ArgumentParser(description="Merge ADIF files")
     parser.add_argument('--problems', '-p', type=str,
                         help="Intermediate problem output .json")
-    parser.add_argument('--output', '-o', type=str,
-                        default="qso_merged.adif")
+    parser.add_argument('--output', '-o', type=str, default="qso_merged.adif",
+                        help="Merged log output .adif")
     parser.add_argument('--minimal', '-m', action='store_true',
                         help="Only output important fields")
     parser.add_argument('--merge-window', type=int, default=MERGE_WINDOW,
                         help="Time window for merging discrepent log entries")
     parser.add_argument('--csv', '-c', type=str,
                         help="WSJT-X compatible .log file")
+    parser.add_argument('--log-level', type=str, default="info",
+                        help="Log level for debugging")
     parser.add_argument('--version', '-v', action='version',
                         version="%(prog)s {version}".format(version=__VERSION__))
     parser.add_argument('input', type=str, nargs="+",
                         help="Input file list")
     args = parser.parse_args()
 
+    numeric_level = getattr(logging, args.log_level.upper(), None)
+    if not isinstance(numeric_level, int):
+        raise ValueError("Invalid log-level: {}".format(args.log_level))
+    logging.basicConfig(format='%(levelname)s: %(message)s', level=numeric_level)
+
     qsos = []
+    malformed = []
     for path in args.input:
+        filename = os.path.basename(path)
         raw, _adif_header = read_adif_file(path)
-        processed = [fixup_qso(qso, os.path.basename(path)) for qso in raw]
-        qsos += processed
+        for qso in raw:
+            try:
+                qsos.append(fixup_qso(qso, filename))
+            except ADIFMalformedQSOError as err:
+                logging.warning("Ignoring QSO: %s", err.message)
+                malformed.append(err.expression)
 
     qsos = merge_qsos(qsos, args.merge_window)
 
     if args.problems:
-        dump_problems(qsos, args.problems)
+        dump_problems(qsos, malformed, args.problems)
 
     if args.output:
-        with open(args.output, "w") as adiffile:
+        # ADIF files are supposed to be ascii, not unicode, unfortunately.
+        with open(args.output, "w", encoding=ADIF_ENCODING) as adiffile:
             adif_write(adiffile, qsos, args.minimal)
 
     if args.csv:
-        with open(args.csv, "w") as csvfile:
+        with open(args.csv, "w", encoding=ADIF_ENCODING) as csvfile:
             csv_write(csvfile, qsos)
 
 
